
# THIS FILE IS GENERATED FROM SigProfilerAssignment SETUP.PY
short_version = '0.0.14'
version = '0.0.14'
Update = '1. Add support for multi-sample segmentation files. 2. Move decomposition plots to SigProfilerAssignment.'
    
    